#include "config.h"
#include "macros.h"
#include <sys/stat.h>
#include "network.h"
#include "readconf.h"
#include "ag.h"
#include "share.h"
#include "filetransfer.h"
#include "ag_fnc.h"
#include "queue.h"
#include "mp3_info.h"
#include "ogg_info.h"

/* sigh */
#define NULLSTR		(unsigned char *)""

unsigned char *ag_loginstr = NULL;


int ag_srv_msg(int msg, const void *data, int len, int prio)
{
	unsigned char *packet;
	int r;

	context;

	if(!data)
		len = 0;

	packet = ag_malloc(5+len);
	if(!packet)
		return -1;

	packet[0] = 0x45;
	packet[1] = 0x5F;
	packet[2] = (len+3) / 256;
	packet[3] = (len+3) % 256;
	packet[4] = msg;

	context;

	if(len)
		memcpy(&packet[5], data, len);
	len += 5;

	/* only send packets if the queue is empty */
	if(queue_is_empty())
	{
		r = send(srv_s, packet, len, 0);
	}
	else
	{
		/* queue it */
		queue_add(packet, len, 0, prio);
		errno = 0;
		return len;
	}

	if(r > 0 && r < len)
	{
		/* partial packet sent */
		queue_add(packet, len, r, prio);
		return r;
	}
	else if(r == -1 && (errno == EAGAIN || errno == EWOULDBLOCK || errno == ENOBUFS))
	{
		queue_add(packet, len, 0, prio);
		errno = 0;
		return len;
	}
	ag_free(packet);

	return r;
}


int ag_magic_response(int len, const unsigned char *data)
{
	unsigned char *ptr,*pkt;
	unsigned char *pass;
	int pkt_len,pass_len=0,r;

	context;

	/* magic + password */
	pass_len = len + strlen(conf.password) - 2;
	pass = ag_malloc(pass_len);
	if(!pass)
		return -1;
	memcpy(pass, data+2, len-2);
	memcpy(pass+len-2, conf.password, strlen(conf.password));

	context;

	/* login packet size */
	pkt_len = 4+2+2+2+2+2+2+2+/*md5*/16;
	pkt_len += strlen(conf.username);
	pkt_len += strlen(AG_CLIENT_VERSION);
	pkt_len += strlen(conf.ip);
	pkt_len += strlen(conf.hostname);
	ptr = pkt = ag_malloc(pkt_len);
	if(!pkt)
	{
		ag_free(pass);
		return -1;
	}

	context;

	/* fill packet */
	AG_PUSH_INT(pkt, 0);					/* unknown */
	AG_PUSH_STR(pkt, conf.username);		/* username */
	AG_PUSH_WORD(pkt, 0);					/* null str? */
	AG_PUSH_STR(pkt, AG_CLIENT_VERSION);	/* version */
	AG_PUSH_STR(pkt, conf.ip);				/* ip address */
	AG_PUSH_STR(pkt, conf.hostname);		/* hostname */
	AG_PUSH_WORD(pkt, 0);					/* # of files sent */
	AG_PUSH_WORD(pkt, 0);					/* # of files received */

	context;
	memcpy(pkt, MD5(pass, pass_len), 16);

	r = ag_srv_msg(AG_MSG_LOGIN, ptr, pkt_len, QPRIO_LOW);
	ag_free(ptr);
	ag_free(pass);

	return r;
}


/* informs the server about what's happening :) */
int ag_send_status()
{
	unsigned char *ptr,*pkt;
	int pkt_len=0;
	int i,r=0;

	/* figure out how much memory we need */
	for(i=0;i<MAX_FT;i++)
	{
		if(ft[i].status != FT_ACTIVE)
			continue;

		pkt_len += 2+ft[i].fileid_len+4+4+4;
	}

	if(pkt_len)
	{
		pkt = ptr = ag_malloc(pkt_len);

		for(i=0;i<MAX_FT;i++)
		{
			if(ft[i].status != FT_ACTIVE)
				continue;

			AG_PUSH_BIN(pkt, ft[i].fileid, ft[i].fileid_len);
			AG_PUSH_INT(pkt, ft[i].speed);
			AG_PUSH_INT(pkt, 0);	/* unknown */
			AG_PUSH_INT(pkt, (int)ftell(ft[i].fd));
		}

		r = ag_srv_msg(AG_MSG_STATUS, ptr, pkt_len, QPRIO_HIGH);
		ag_free(ptr);
	}
	else
	{
		r = ag_srv_msg(AG_MSG_STATUS, NULL, 0, QPRIO_HIGH);
	}

	return r;
}


int ag_send_new_share(struct sharestruct *s, int *cost)
{
	unsigned char file[MAX_PATH];
	unsigned char *ptr,*pkt;
	int pkt_len=0,r;
	int16 type=0x0003;

	context;

	(*cost)++;

	if(!s->size)
	{
		NOTICE("Tried to send empty file in ag_send_new_share()");
		return 1;
	}

	snprintf(file, MAX_PATH, "%s/%s", s->path, s->file);

	/* free() some stuff */
	share_wash(s);

	switch(get_file_type(s->file))
	{
	case SHARE_TYPE_MP3:
		mp3_fill_info(s, file, s->size);
		break;
#ifdef USE_VORBIS
	case SHARE_TYPE_OGG:
		ogg_fill_info(s, file, 1);
		break;
#endif
	}

	(*cost) += 34;		/* looks ok */

	if(share_get_md5_hash(s))
		(*cost) += 65;	/* read successful - expensive */
	else
		(*cost)++;		/* no data read - cheap */

	pkt_len = 2+4+2+2+4+2+4+16+4;
	pkt_len += s->file ? strlen(s->file) : 0;
	pkt_len += s->path ? strlen(s->path) : 0;
	if(s->have_id3v1)
	{
		type |= 0x0004;
		pkt_len += 2+2+2;
		pkt_len += s->id3v1_title ? strlen(s->id3v1_title) : 0;
		pkt_len += s->id3v1_artist ? strlen(s->id3v1_artist) : 0;
		pkt_len += s->id3v1_album ? strlen(s->id3v1_album) : 0;
	}
	if(s->have_id3v2)
	{
		type |= 0x0008;
		pkt_len += 2+2+2+2+/* XAG tag */2+3;
		pkt_len += s->id3v2_title ? strlen(s->id3v2_title) : 0;
		pkt_len += s->id3v2_artist ? strlen(s->id3v2_artist) : 0;
		pkt_len += s->id3v2_album ? strlen(s->id3v2_album) : 0;
	}

	ptr = pkt = ag_malloc(pkt_len);
	if(!pkt) return -1;

	context;

	/* fill packet */
	AG_PUSH_WORD(pkt, type);				/* unknown */
	AG_PUSH_INT(pkt, s->id);				/* local id */
	AG_PUSH_STR(pkt, s->file?s->file:NULLSTR);	/* filename */
	AG_PUSH_STR(pkt, s->path?s->path:NULLSTR);	/* path */
	AG_PUSH_INT(pkt, s->size);				/* filesize */
	AG_PUSH_WORD(pkt, s->bitrate);			/* bitrate */
	AG_PUSH_INT(pkt, s->seclength);			/* length of song */
	AG_PUSH_INT(pkt, s->lastmod);			/* last modified */
	AG_PUSH_RBIN(pkt, s->file_hash, 16);	/* md5 */
	/* ID3v1 tag */
	if(s->have_id3v1)
	{
		AG_PUSH_STR(pkt, s->id3v1_title ? s->id3v1_title : NULLSTR);
		AG_PUSH_STR(pkt, s->id3v1_artist ? s->id3v1_artist : NULLSTR);
		AG_PUSH_STR(pkt, s->id3v1_album ? s->id3v1_album : NULLSTR);
	}
	/* ID3v2 tag */
	if(s->have_id3v2)
	{
		AG_PUSH_STR(pkt, s->id3v2_title ? s->id3v2_title : NULLSTR);
		AG_PUSH_STR(pkt, s->id3v2_artist ? s->id3v2_artist : NULLSTR);
		AG_PUSH_STR(pkt, s->id3v2_album ? s->id3v2_album : NULLSTR);
		AG_PUSH_WORD(pkt, s->id3v2_track);
		AG_PUSH_STR(pkt, "XAG");
	}

	r = ag_srv_msg(AG_MSG_NEWSHAREDFILE, ptr, pkt_len, QPRIO_LOW);
	ag_free(ptr);

	return r;
}


int ag_send_upd_share(struct sharestruct *s, int *cost)
{
	struct stat st;
	unsigned char file[MAX_PATH];

	context;

	ag_send_old_share(s, cost);
	ag_send_del_share(s, cost);

	/* free some stuff */
	share_wash(s);

	snprintf(file, sizeof(file), "%s/%s", s->path, s->file);
	if(stat(file, &st))
	{
		/* stat failed.. delete the share! */
		NOTICE("Can't stat '%s': %s", file, ERRMSG());
		share_del(s);
		return 1;
	}
	s->size = st.st_size;
	s->lastmod = st.st_mtime;

	upd_shares--;
	new_shares++;
	s->status = SHARE_STAT_NEW;

	ag_send_new_share(s, cost);

	return 1;
}


int ag_send_old_share(struct sharestruct *s, int *cost)
{
	unsigned char *ptr,*pkt;
	int pkt_len=0,r;

	context;

	(*cost)++;

	if(!s->size)
	{
		NOTICE("Tried to send empty file in ag_send_old_share()");
		return 1;
	}

	pkt_len = 4+2+s->fileid_len+4+4+4+2+4;
	ptr = pkt = ag_malloc(pkt_len);
	if(!pkt) return -1;

	/* fill packet */
	AG_PUSH_INT(pkt, s->id);				/* local id */
	AG_PUSH_BIN(pkt, s->fileid, s->fileid_len);
	AG_PUSH_INT(pkt, s->server_id);			/* server file id */
	AG_PUSH_INT(pkt, s->song_id);			/* song id? */
	AG_PUSH_INT(pkt, s->size);				/* filesize */
	AG_PUSH_WORD(pkt, s->bitrate);			/* bitrate */
	AG_PUSH_INT(pkt, s->lastmod);			/* last modified */

	r = ag_srv_msg(AG_MSG_OLDSHAREDFILE, ptr, pkt_len, QPRIO_LOW);
	ag_free(ptr);

	return r;
}



int ag_send_del_share(struct sharestruct *s, int *cost)
{
	unsigned char *ptr,*pkt;
	int pkt_len=0,r;

	context;

	(*cost)++;

	save_share_db = 1;

	if(!s->size)
	{
		NOTICE("Tried to send empty file in ag_send_del_share()");
		return 1;
	}

	pkt_len = 4+4;
	ptr = pkt = ag_malloc(pkt_len);
	if(!pkt) return -1;

	/* fill packet */
	AG_PUSH_INT(pkt, s->id);		/* local id */
	AG_PUSH_INT(pkt, s->size);		/* filesize */

	r = ag_srv_msg(AG_MSG_FILEDELETED, ptr, pkt_len, QPRIO_LOW);
	ag_free(ptr);

	return r;
}


void ag_share_reply(const unsigned char *data, int len)
{
	int id;
	struct sharestruct *share;

	context;

	AG_POP_INT(id, data, len);	/* local id */
	if(len == -1)
		return;

	share = share_find_by_id(id);
	if(!share)
	{
		NOTICE("Server returned unknown local id!");
		return;
	}

	share->fileid_len = AG_WORD(data);
	AG_POP_STR(share->fileid, data, len);
	AG_POP_INT(share->server_id, data, len);
	
	share->status = SHARE_STAT_OLD;
	old_shares++;
	new_shares--;
	AG_POP_INT(share->song_id, data, len);
	/* 1 int left in packet: filesize */

	if(share->is_download)
	{
		/* delayed download_end trigger */
		event_action("download_end", 4, share->path, share->file,
				float2str((float)share->download_speed/1024), int2str(share->song_id));
		share->is_download = 0;
	}

	shared++;
	save_share_db = 1;
}


/* just replies to delete requests - doesn't really delete any file */
void ag_delete_reply(const unsigned char *data, int len)
{
	unsigned char *fileid,*ptr,*pkt;
	struct sharestruct *share;
	int pkt_len,fileid_len;

	context;

	if(len >= 2)
	{
		fileid_len = AG_WORD(data);
		AG_POP_STR(fileid, data, len);
		if(len == -1)
		{
			if(fileid)
				ag_free(fileid);
			return;
		}

		share = share_find_by_fileid(fileid, fileid_len);
		if(share)
		{
			/* trigger event */
			event_action("delete", 2, share->path, share->file);
		}

		if(fileid)
		{
			pkt_len = 2+fileid_len+1;
			pkt = ptr = ag_malloc(pkt_len);
			if(!ptr)
			{
				ag_free(fileid);
				return;
			}
			AG_PUSH_BIN(ptr, fileid, fileid_len);
			AG_PUSH_BYTE(ptr, 1);
			ag_srv_msg(AG_MSG_DELETEREPLY, pkt, pkt_len, QPRIO_LOW);
			ag_free(pkt);
			ag_free(fileid);
		}
	}
}


void ag_playsonglist_reply(const unsigned char *data, int len)
{
	unsigned char *opstr;
	int op, local_id;
	struct sharestruct *share;

	if(len < 12)	/* must have at least 3 ints */
		return;

	AG_POP_INT(op, data, len);
	/* INT: count - number of "local_id" and "unknown" */
	data += 4; len -= 4;
	AG_POP_INT(local_id, data, len);
	/* INT: unknown*/

	share = share_find_by_id(local_id);
	if(!share)
	{
		NOTICE("Server returned unknown local id!");
		return;
	}

	switch(op)
	{
	case 1:
		opstr = "play";
		break;
	case 2:
		opstr = "queue";
		break;
	default:
		opstr = "unknown";
		break;
	}

	/* trigger event */
	event_action("play_file", 3, opstr, share->path, share->file);
}


void ag_stoptransfer_reply(const unsigned char *data, int len)
{
	unsigned char *fileid;
	int fileid_len, i;

	if(len >= 2)
	{
		fileid_len = AG_WORD(data);
		AG_POP_STR(fileid, data, len);
		if(fileid) {
			i = ft_find_by_fileid(fileid, fileid_len);
			if(i != -1)
			{
				NOTICE("Stopping transfer of '%s' as requested by the server.", ft[i].file);
				ft_send_state(AG_FT_STATE_STOPPED, ft[i].fileid, ft[i].fileid_len);
			}
			ag_free(fileid);
		}
	}
}


void ag_loggedin_reply(const unsigned char *data, int len)
{
	if(status != AG_STAT_LOGGED_IN)
	{
		NOTICE("Logged in!");

		/* trigger event */
		if(len >= 2)
		{
			AG_POP_STR(ag_loginstr, data, len);
			if(ag_loginstr)
			{
				event_action("login", 1, ag_loginstr);
			}
			else
			{
				event_action("login", 1, "");
			}
		}
	}
	status = AG_STAT_LOGGED_IN;
}


unsigned char *ag_msg_type(int msg)
{
	switch(msg) {
	case AG_MSG_LOGIN:
		return "AG_MSG_LOGIN";
	case AG_MSG_LOGOUT:
		return "AG_MSG_LOGOUT";
	case AG_MSG_FILEDELETED:
		return "AG_MSG_FILEDELETED";
	case AG_MSG_STATUS:
		return "AG_MSG_STATUS";
	case AG_MSG_FILETRANSFERSTATE:
		return "AG_MSG_FILETRANSFERSTATE";
	case AG_MSG_READYTOSENDSHARES:
		return "AG_MSG_READYTOSENDSHARES";
	case AG_MSG_SENDSHARESSTANDBY:
		return "AG_MSG_SENDSHARESSTANDBY";
	case AG_MSG_SENDSHARES:
		return "AG_MSG_SENDSHARES";
	case AG_MSG_FILETRANSFER:
		return "AG_MSG_FILETRANSFER";
	case AG_MSG_STOPFILETRANSFER:
		return "AG_MSG_STOPFILETRANSFER";
	case AG_MSG_NEWSHAREREPLY:
		return "AG_MSG_NEWSHAREREPLY";
	case AG_MSG_LOGINERROR:
		return "AG_MSG_LOGINERROR";
	case AG_MSG_PLAYFILE:
		return "AG_MSG_PLAYFILE";
	case AG_MSG_ACCOUNTEXISTENCE:
		return "AG_MSG_ACCOUNTEXISTENCE";
	case AG_MSG_OLDSHAREDFILE:
		return "AG_MSG_OLDSHAREDFILE";
	case AG_MSG_NEWSHAREDFILE:
		return "AG_MSG_NEWSHAREDFILE";
	case AG_MSG_DELETEREPLY:
		return "AG_MSG_DELETEREPLY";
	case AG_MSG_NEWACCOUNT:
		return "AG_MSG_NEWACCOUNT";
	case AG_MSG_MAGICREQUEST:
		return "AG_MSG_MAGICREQUEST";
	case AG_MSG_NEWACCOUNTREPLY:
		return "AG_MSG_NEWACCOUNTREPLY";
	case AG_MSG_ACCOUNTEXISTENCEREPLY:
		return "AG_MSG_ACCOUNTEXISTENCEREPLY";
	case AG_MSG_DELETEREQUEST:
		return "AG_MSG_DELETEREQUEST";
	case AG_MSG_PLAYSONGLIST:
		return "AG_MSG_PLAYSONGLIST";
	case AG_MSG_MAGICRESPONSE:
		return "AG_MSG_MAGICRESPONSE";
	case AG_MSG_LOGGED_IN:
		return "AG_MSG_LOGGED_IN";
	default:
		return "AG_MSG_UNKNOWN";
	}
}


/* handles a server packet */
int ag_handle_packet(int len, unsigned char msg, const unsigned char *data)
{
	static int sss=0;
	unsigned char *errmsg;
	int tmp;

	context;

	DEBUG("Received packet %d('%s') length %d bytes.", msg, ag_msg_type(msg), len);

	switch(msg)
	{
	case AG_MSG_MAGICRESPONSE:
		return ag_magic_response(len, data);

	case AG_MSG_LOGINERROR:
		if(len < 4)
		{
			NOTICE("Login packet too short!");
			return -1;
		}

		tmp = AG_INT(data);
		if(tmp == AG_LOGINERR_INVALID) {
			errmsg = "Invalid username or password";
		} else if(tmp == AG_LOGINERR_VERSION) {
			errmsg = "Bad client version";
		} else if(tmp == AG_LOGINERR_LOGGEDIN) {
			running = 0;
			NOTICE("Another user has logged in on this account. Exiting...");
			errmsg = "Another user has logged in on this account";
		} else if(tmp == AG_LOGINERR_GOLDEXPIRED) {
			errmsg = "Gold account has expired";
		} else if(tmp == AG_LOGINERR_NOTGOLD) {
			errmsg = "Account is not Gold";
		} else {
			errmsg = "Unknown error";
		}
		if(tmp != AG_LOGINERR_LOGGEDIN)
			NOTICE("Login error! %s.", errmsg);
		event_action("login_error", 1, errmsg);
		return -1;

	case AG_MSG_LOGGED_IN:
		ag_loggedin_reply(data, len);
		sss = 0;
		return 0;

	/* server wants us to prepare to send shares */
	case AG_MSG_SENDSHARESSTANDBY:
		/* we are always ready to send shares. */
		if(!sss)
		{
			NOTICE("Ready to send shares...");
			NOTICE("The server is about to run a proxy test, please ignore all errors.");
		}
		ag_srv_msg(AG_MSG_READYTOSENDSHARES, NULL, 0, QPRIO_HIGH);
		sss = 1;
		break;

	case AG_MSG_SENDSHARES:
		NOTICE("Proxy test completed.");
		NOTICE("Sending shares...");
		status = AG_STAT_SEND_SHARES;
		break;

	case AG_MSG_NEWSHAREREPLY:
		ag_share_reply(data, len);
		break;

	case AG_MSG_STOPFILETRANSFER:
		ag_stoptransfer_reply(data, len);
		break;

	case AG_MSG_DELETEREQUEST:
		ag_delete_reply(data, len);
		break;

	case AG_MSG_PLAYSONGLIST:
		ag_playsonglist_reply(data, len);
		break;

	default:
		NOTICE("Received unhandled packet %d('%s') length %d bytes.", msg, ag_msg_type(msg), len);
	}

	return -1;
}

