#ifndef _FILETRANSFER_H
#define _FILETRANSFER_H

#define FT_UNUSED		0	/* not in use */
#define FT_CONNECTING	1	/* connecting */
#define FT_LISTENING	2	/* waiting for connections */
#define FT_WAIT			3	/* waiting for response/request */
#define FT_ACTIVE		4	/* transferring / receiving */

/* file transfer states */
#define AG_FT_STATE_COMPLETE	0
#define AG_FT_STATE_ERROR		1
#define AG_FT_STATE_SETUP		3
#define AG_FT_STATE_NOCONN		4	/* could not connect */
#define AG_FT_STATE_CLOSED		5
#define AG_FT_STATE_ERRBIND		7
#define AG_FT_STATE_STOPPED		8

#define AG_FT_SEND			0
#define AG_FT_RECV			1

#define AG_FT_LISTEN		0
#define AG_FT_CONNECT		1

#define AG_FD_READ			0
#define AG_FD_WRITE			1


struct struct_ft {
	int s;					/* socket */
	int status;				/* status */
	struct timeval start;	/* when the transfer started */
	struct timeval timeout;	/* timeout */
	FILE *fd;

	int direction;			/* are we sending or receiving? */
	int connecting;			/* are we connecting or listening? */
	int fdset;				/* which fd_set? (read or write ;) */
	int fileid_len;			/* file id length */
	unsigned char *fileid;
	unsigned char *ip;
	int port;
	int resume_pos;			/* where we resumed */
	int filesize;			/* filesize */
	/*int position;*/			/* position - use ftell(ft[id].fd) ! */
	unsigned char *file;
	unsigned char *ufile;	/* upload filename (full path) */

	unsigned char *buf;		/* read buffer */

	/* transfer speed stuff */
	struct timeval lastspd;	/* last speed check */
	int lastspd_pos;		/* last speed check position */
	int speed;				/* speed in bytes per second - updated every 4 seconds */

	/* bandwidth  stuff */
	int bw;					/* bandwidth left */
	int bw_last;			/* how much bandwidth we got the last time  */
};

extern struct struct_ft ft[MAX_FT];
extern int ft_uploads,ft_downloads;
extern fd_set ft_read_fd,ft_write_fd;

void ft_init();
int ft_alloc();
void ft_free(int id);
void ft_close_conn(int id);
void ft_close_all();
void ft_update_speed(int id);
int ft_max_fd();
void ft_new_fd(int fd);
void ft_del_fd(int fd);
int ft_find_by_fileid(unsigned char *fileid, int fileid_len);
float ft_avg_speed(int id);
int ft_fillstruct(int id, int len, unsigned char *data);
int ft_send_state(int state, const unsigned char *fileid, int fileid_len);
void ag_do_filetransfer(int srv_s, int conn, int direction, const unsigned char *ip,
						int port, const unsigned char *filename, int filesize,
						const unsigned char *file_id, int file_id_len);
FILE *ft_fopen(unsigned char *file, int rw);
void ft_rename(int id);
int ft_recv_request(int id, int len, unsigned char *data);
int ft_recv_response(int id, int len, unsigned char *data);
int ft_send_request(int id, int resume_pos);
int ft_send_response(int id);

#endif
