#ifndef _MACROS_H
#define _MACROS_H

#define DEBUG(fmt, args...)		if(debug) debug_handler(__FILE__, __FUNCTION__, __LINE__, fmt, ## args)
#define ERRMSG()				errmsg_handler(__FUNCTION__, __LINE__)
#define NOTICE(fmt, args...)	notice_handler(__FUNCTION__, fmt, ## args)

#define AG_VALID_HDR(s)			(s[0] == 0x45 && s[1] == 0x5F)
#define AG_VALID_REDIR(s)		(s[2] == 0xD0 && s[3] == 0xD5)
#define AG_VALID_PHDR(s)		AG_VALID_REDIR(s)		/* peer msg header - same as redir */

#define AG_INT(ptr)				ag_ptr_int32(ptr)
#define AG_WORD(ptr)			ag_ptr_int16(ptr)

#define AG_PUSH_BIN(dst,src,len)	memcpy(dst, ag_int16_ptr(len), 2); memcpy(dst+2, src, len); dst += len+2
#define AG_PUSH_STR(dst,src)		AG_PUSH_BIN(dst,src,strlen(src))
#define AG_PUSH_INT(dst,src)		memcpy(dst, ag_int32_ptr(src), 4); dst += 4
#define AG_PUSH_WORD(dst,src)		memcpy(dst, ag_int16_ptr(src), 2); dst += 2
#define AG_PUSH_BYTE(dst,src)		*dst++ = src
#define AG_PUSH_RBIN(dst,src,len)	memcpy(dst, src, len); dst += len

#define AG_POP_STR(dst,src,len)		if(len >= 2) { dst = ag_pop_str_handler(src,&len); src += AG_WORD(src)+2; }	else { NOTICE("AG_POP_STR underflow in %s", __FUNCTION__); len = -1; dst = NULL; }
#define AG_POP_INT(dst,src,len)		if(len >= 4) { dst = AG_INT(src); src += 4; len -= 4; } else { NOTICE("AG_POP_INT underflow in %s", __FUNCTION__); len = -1; dst = 0; }
#define AG_POP_WORD(dst,src,len)	if(len >= 2) { dst = AG_WORD(src); src += 2; len -= 2; } else { NOTICE("AG_POP_WORD underflow in %s", __FUNCTION__); len = -1; dst = 0; }
#define AG_POP_BYTE(dst,src,len)	if(len >= 1) { dst = *src++; len--; } else { NOTICE("AG_POP_BYTE underflow in %s", __FUNCTION__); dst = 0; }

#ifdef USE_CONTEXTS
#	define context snprintf(_context[_ncontext++%_NCX], sizeof(_context[0]), "%s/%s:%d", __FILE__, __FUNCTION__, __LINE__)
#else
#	define context
#endif

void debug_handler(const unsigned char *file, const unsigned char *function, int line, const unsigned char *format, ...);
unsigned char *errmsg_handler(const unsigned char *function, int line);
void notice_handler(const unsigned char *function, const unsigned char *format, ...);
unsigned char *ag_pop_str_handler(const unsigned char *src, int *len);

/* usage: DIFF_MSEC(last, cur) */
#define DIFF_MSEC(b,a) \
	( (a.tv_sec*1000 + a.tv_usec/1000) - (b.tv_sec*1000 + b.tv_usec/1000) )

#define LOG_ENABLE()	logfile = _logfile
#define LOG_DISABLE()	logfile = NULL

extern int debug;
extern int background;
extern FILE *logfile,*_logfile;
extern struct timeval uptime;
#ifdef USE_CONTEXTS
#define _NCX 20
extern int _ncontext;
extern unsigned char _context[_NCX][128];
#endif


#endif
