#include "config.h"
#include "macros.h"
#include "network.h"
#include <stdarg.h>

unsigned char *time_string()
{
	static unsigned char str[8] = "[XX:XX]";
	time_t ct = time(NULL);
	struct tm *t = localtime(&ct);

	if(t)
		snprintf(str, sizeof(str), "[%02d:%02d]", t->tm_hour, t->tm_min);

	return str;
}


void debug_handler(const unsigned char *file, const unsigned char *function, int line, const unsigned char *format, ...)
{
	va_list va;

	fprintf(stderr, "%s (%s/%s:%d) DEBUG: * ", time_string(), file, function, line);
	va_start(va, format);
	vfprintf(stderr, format, va);
	va_end(va);
	fprintf(stderr, "\n");
}


unsigned char *errmsg_handler(const unsigned char *function, int line)
{
	static unsigned char str[128];
	unsigned char *errstr = strerror(errno);

	if(debug)
		snprintf(str, sizeof(str), "%s (%d) (%s:%d)",
			errstr?errstr:(unsigned char *)"Unknown error",
			errno,
			function, line);
	else
		snprintf(str, sizeof(str), "%s",
			errstr?errstr:(unsigned char *)"Unknown error");

	return str;
}


void notice_handler(const unsigned char *function, const unsigned char *format, ...)
{
	static unsigned char str[384];
	va_list va;

	va_start(va, format);
	vsnprintf(str, sizeof(str), format, va);
	va_end(va);

	if(!background)
		printf("%s %s\n", time_string(), str);

	if(logfile)
	{
		fprintf(logfile, "%s %s\n", time_string(), str);
		fflush(logfile);
	}
}


unsigned char *ag_pop_str_handler(const unsigned char *src, int *len)
{
	unsigned char *dst;
	int l = AG_WORD(src);

	if((l+2) > *len)
	{
		NOTICE("Invalid string length in packet!");
		l -= 2;
		return NULL;
	}

	dst = ag_malloc(l+1);
	if(!dst) return NULL;

	*len -= l+2;

	memcpy(dst, src+2, l);
	dst[l] = 0;

	return dst;
}

