#include "config.h"
#include "macros.h"
#include "share.h"
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>
#include "crc.h"
#include "share_helper.h"

struct sharestruct *sharelist=NULL,*curshare=NULL;
int old_shares=0,new_shares=0,upd_shares=0,del_shares=0,shared=0,save_share_db=0;
unsigned char *curpath=NULL;


void share_status()
{
	NOTICE("Share status: %d old, %d new, %d updated, %d deleted.",
			old_shares, new_shares, upd_shares, del_shares);
}


struct sharestruct *share_find_by_id(int id)
{
	struct sharestruct *share=sharelist;

	for(;share;share=share->next)
		if(share->id == id)
			return share;

	return NULL;
}


struct sharestruct *share_find_by_fileid(unsigned char *fileid, int fileid_len)
{
	struct sharestruct *share=sharelist;

	for(;share;share=share->next)
		if(share->fileid_len == fileid_len)
			if(share->fileid && !memcmp(share->fileid, fileid, fileid_len))
				return share;

	return NULL;
}


void share_interval_save_db()
{
	static struct timeval tv,ltv;
	static int first_call=1;

	if(first_call)
	{
		gettimeofday(&ltv, NULL);
		first_call = 0;
	}

	if(save_share_db)
	{
		gettimeofday(&tv, NULL);
		if(tv.tv_sec > ltv.tv_sec + DBSAVE_MININT)
		{
			share_save_db();
			gettimeofday(&ltv, NULL);
			save_share_db = 0;
		}
	}
}


/* make sure s->size is set before calling */
int share_get_md5_hash(struct sharestruct *s)
{
	static unsigned char buf[32768],*ptr;
	unsigned char bad_md5[16]={0};
	unsigned char file[MAX_PATH];
	int p=0,i;
	FILE *fd;

	context;

	if(!s)
	{
		NOTICE("share_get_md5_hash() called with NULL arg!");
		return 0;
	}

	/* this doesnt hurt */
	if(s->file_hash)
		ag_free(s->file_hash);
	context;

	s->file_hash = ag_malloc(16);
	if(!s->file_hash)
		return 0;

	if(snprintf(file, MAX_PATH, "%s/%s", s->path, s->file) == -1)
		goto md5_hash_err;

	fd = fopen(file, "r");
	if(!fd)
		goto md5_hash_err;

	for(i=0x100000;i>=0x8000;i/=2)
	{
		if(s->size > i)
		{
			p = i;
			break;
		}
	}
	/* file too small to generate md5 sum */
	if(!p)
		goto md5_hash_err;

	ptr = buf;
	for(i=0;i<p;i+=(p/32))
	{
		fseek(fd, i, SEEK_SET);
		fread(ptr, 1, 1024, fd);
		ptr += 1024;
	}

	fclose(fd);

	memcpy(s->file_hash, MD5(buf, sizeof(buf)), 16);

	/* hmm */
	DEBUG("MD5 hash of file content: %02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X%02X",
		s->file_hash[0], s->file_hash[1], s->file_hash[2], s->file_hash[3],
		s->file_hash[4], s->file_hash[5], s->file_hash[6], s->file_hash[7],
		s->file_hash[8], s->file_hash[9], s->file_hash[10], s->file_hash[11],
		s->file_hash[12], s->file_hash[13], s->file_hash[14], s->file_hash[15]);

	return 1;

md5_hash_err:;
	DEBUG("could not caluclate MD5 hash for file %s/%s", s->path, s->file);
	memcpy(s->file_hash, bad_md5, sizeof(bad_md5));
	context;
	return 0;
}


int path_is_shared(unsigned char *path)
{
	struct confshare *s = conf.sharelist;
	int len = strlen(s->path);

	while(s)
	{
		if(!strncmp(s->path, path, len))
			return 1;
		s = s->next;
	}

	return 0;
}


struct sharestruct *share_new()
{
	static int shareid=0;

	/* add filename to list */
	if(!sharelist)
	{
		context;
		sharelist = curshare = ag_malloc(sizeof(struct sharestruct));
		if(!curshare) return NULL;

		memset(curshare, 0, sizeof(struct sharestruct));
	}
	else
	{
		context;
		curshare->next = ag_malloc(sizeof(struct sharestruct));
		if(!curshare->next) return NULL;

		memset(curshare->next, 0, sizeof(struct sharestruct));

		curshare->next->prev = curshare;
		curshare = curshare->next;
	}

	curshare->id = shareid++;

	return curshare;
}


void share_wash(struct sharestruct *s)
{
	context;

	DEBUG("washing share");
	if(s->file_hash)
		ag_free(s->file_hash);
	if(s->fileid)
		ag_free(s->fileid);
	if(s->id3v1_title)
		ag_free(s->id3v1_title);
	if(s->id3v1_artist)
		ag_free(s->id3v1_artist);
	if(s->id3v1_album)
		ag_free(s->id3v1_album);
	if(s->id3v2_title)
		ag_free(s->id3v2_title);
	if(s->id3v2_artist)
		ag_free(s->id3v2_artist);
	if(s->id3v2_album)
		ag_free(s->id3v2_album);

	return;
}


void share_del(struct sharestruct *share)
{
	context;

	DEBUG("deleting share");

	/* remove share from the list */
	if(share->prev)
		share->prev->next = share->next;

	if(share->next)
		share->next->prev = share->prev;

	if(!share->next)	/* share was at the end of the chain */
		curshare = share->prev;

	if(!share->prev)	/* share was at the start of the chain */
		sharelist = share->next;

	if(share->path)
		ag_free(share->path);
	if(share->file)
		ag_free(share->file);
	if(share->file_hash)
		ag_free(share->file_hash);
	if(share->fileid)
		ag_free(share->fileid);
	if(share->id3v1_title)
		ag_free(share->id3v1_title);
	if(share->id3v1_artist)
		ag_free(share->id3v1_artist);
	if(share->id3v1_album)
		ag_free(share->id3v1_album);
	if(share->id3v2_title)
		ag_free(share->id3v2_title);
	if(share->id3v2_artist)
		ag_free(share->id3v2_artist);
	if(share->id3v2_album)
		ag_free(share->id3v2_album);

	ag_free(share);

	return;
}


unsigned long share_make_hash(unsigned char *path, unsigned char *file)
{
	unsigned long hash;
	unsigned char *str;

	context;

	str = ag_malloc(strlen(path)+strlen(file)+1);
	if(!str) return 0;

	strcpy(str, path);
	strcat(str, file);

	hash = crc32(str);
	ag_free(str);

	return hash;
}


struct sharestruct *is_shared(unsigned char *path, unsigned char *file)
{
	struct sharestruct *share=sharelist;
	unsigned long hash;

	context;

	/* no this is not a hash table, but it's fast enough. :P */

	hash = share_make_hash(path, file);
	while(share)
	{
		if(share->hash == hash)
		{
			/* make sure it's a 100% accurate match */
			if(!strcmp(share->file, file) && !strcmp(share->path, path))
				return share;
		}
		share = share->next;
	}

	return 0;
}


int get_file_type(unsigned char *file)
{
	int i;

	context;

	i = strlen(file);
	if(i >= 4)
	{
		if(!strcasecmp(&file[i-4], ".mp3"))
			return SHARE_TYPE_MP3;
#ifdef USE_VORBIS
		if(!strcasecmp(&file[i-4], ".ogg"))
			return SHARE_TYPE_OGG;
#endif
	}

	return SHARE_TYPE_UNKNOWN;
}



int share_save_db()
{
	struct sharestruct *share=sharelist;
	unsigned char hdr[12];
	int err=0;
	int rollback=0;
	FILE *fd;

	if(!share)
		return 1;

	context;

	NOTICE("Saving share db...");

	fd = fopen(etc_file("sharedb"), "w");
	if(!fd)
	{
		NOTICE("Could not open '%s': %s", etc_file("sharedb"), ERRMSG());
		return 0;
	}

	memset(hdr, 0, sizeof(hdr));
	memcpy(hdr, "FagsDB0x0006" , sizeof(hdr));
	fwrite(hdr, sizeof(hdr), 1, fd);

	context;
	while(share && !err)
	{
		rollback = ftell(fd);

		FWRITE_UINT32(fd, &err, share->hash);
		FWRITE_INT(fd, &err, share->status);

		FWRITE_STR(fd, &err, share->path);
		FWRITE_STR(fd, &err, share->file);
		FWRITE_BIN(fd, &err, share->file_hash, sizeof(share->file_hash));
		FWRITE_INT(fd, &err, share->size);
		FWRITE_INT(fd, &err, share->lastmod);

		FWRITE_BIN(fd, &err, share->fileid, share->fileid_len);
		FWRITE_INT(fd, &err, share->server_id);
		FWRITE_INT(fd, &err, share->song_id);

		FWRITE_SHORT(fd, &err, share->bitrate);
		FWRITE_INT(fd, &err, share->seclength);

		FWRITE_BYTE(fd, &err, share->have_id3v1);
		if(share->have_id3v1)
		{
			context;
			FWRITE_STR(fd, &err, share->id3v1_title);
			FWRITE_STR(fd, &err, share->id3v1_artist);
			FWRITE_STR(fd, &err, share->id3v1_album);
		}

		FWRITE_BYTE(fd, &err, share->have_id3v2);
		if(share->have_id3v2)
		{
			context;
			FWRITE_STR(fd, &err, share->id3v2_title);
			FWRITE_STR(fd, &err, share->id3v2_artist);
			FWRITE_STR(fd, &err, share->id3v2_album);
			FWRITE_SHORT(fd, &err, share->id3v2_track);
		}

		DEBUG("Wrote share to database: %s/%s", share->path, share->file);

		share = share->next;
	}

	if(err)
		NOTICE("Error while writing sharedb: %s", ERRMSG());

	fclose(fd);

	if(err)
		truncate(etc_file("sharedb"), rollback);

	return !err;
}


int share_load_db()
{
	struct sharestruct *share=NULL;
	unsigned char hdr[12],file[MAX_PATH];
	struct stat st;
	int db_version;
	int err=0,ret,filesize;
	FILE *fd;

	context;

	NOTICE("Reading share db...");

	if(sharelist)
	{
		DEBUG("The sharelist should be empty when calling this function.");
		return 0;
	}

	fd = fopen(etc_file("sharedb"), "r");
	if(!fd)
	{
		if(errno != ENOENT)
		{
			NOTICE("Could not open '%s': %s", etc_file("sharedb"), ERRMSG());
		}

		return 1;
	}

	context;

	fseek(fd, 0, SEEK_END);
	filesize = ftell(fd);
	fseek(fd, 0, SEEK_SET);

	fread(hdr, sizeof(hdr), 1, fd);
	if(!strncmp("FagsDB0x0001", hdr, sizeof(hdr)))
		db_version = 1;
	else if(!strncmp("FagsDB0x0002", hdr, sizeof(hdr)))
		db_version = 2;
	else if(!strncmp("FagsDB0x0003", hdr, sizeof(hdr)))
		db_version = 3;
	else if(!strncmp("FagsDB0x0004", hdr, sizeof(hdr)))
		db_version = 4;
	else if(!strncmp("FagsDB0x0005", hdr, sizeof(hdr)))
		db_version = 5;
	else if(!strncmp("FagsDB0x0006", hdr, sizeof(hdr)))
		db_version = 6;
	else
	{
		NOTICE("Unknown sharedb version!");
		fclose(fd);
		return 0;
	}

	if(db_version == 1)
	{
		/* too many changes */
		NOTICE("Detected old sharedb version 1...");
		fclose(fd);
		return 1;
	}
	if(db_version == 2)
	{
		/* MD5 hashes were not calculated correctly in the devel version */
		NOTICE("Detected devel sharedb version 2...");
		fclose(fd);
		return 1;
	}
	if(db_version == 3)
	{
		NOTICE("Detected sharedb version 3...");
		NOTICE("Due to a 'new share' packet bug in previous versions,");
		NOTICE("all shares must resent as 'new shares' to fix the problem.");
		fclose(fd);
		return 1;
	}
	if(db_version == 4)
	{
		/* new mp3 id3 tag reader */
		NOTICE("Detected devel sharedb version 4...");
		fclose(fd);
		return 1;
	}
	if(db_version == 5)
	{
		/* bitrate and songlength calculated incorrectly :/ */
		NOTICE("Detected sharedb version 5...");
		fclose(fd);
		return 1;
	}

	while(!err)
	{
		context;

		if(ftell(fd) >= filesize)
			break;

		share = share_new();
		if(!share)
		{
			NOTICE("Error allocating new share: %s", ERRMSG());
			break;
		}

		share->hash = FREAD_UINT32(fd, &err);
		share->status = FREAD_INT(fd, &err);

		share->path = FREAD_STR(fd, &err);
		share->file = FREAD_STR(fd, &err);
		share->file_hash = FREAD_BIN(fd, &err, NULL);
		share->size = FREAD_INT(fd, &err);
		share->lastmod = FREAD_INT(fd, &err);

		share->fileid = FREAD_BIN(fd, &err, &share->fileid_len);
		share->server_id = FREAD_INT(fd, &err);
		share->song_id = FREAD_INT(fd, &err);

		share->bitrate = FREAD_SHORT(fd, &err);
		share->seclength = FREAD_INT(fd, &err);

		share->have_id3v1 = FREAD_BYTE(fd, &err);
		if(share->have_id3v1)
		{
			share->id3v1_title = FREAD_STR(fd, &err);
			share->id3v1_artist = FREAD_STR(fd, &err);
			share->id3v1_album = FREAD_STR(fd, &err);
		}

		share->have_id3v2 = FREAD_BYTE(fd, &err);
		if(share->have_id3v2)
		{
			share->id3v2_title = FREAD_STR(fd, &err);
			share->id3v2_artist = FREAD_STR(fd, &err);
			share->id3v2_album = FREAD_STR(fd, &err);
			share->id3v2_track = FREAD_SHORT(fd, &err);
		}


		snprintf(file, sizeof(file), "%s/%s", share->path, share->file);
		ret = stat(file, &st);
		if(ret || !path_is_shared(share->path) || st.st_size == 0)
		{
			/* file has either been deleted, is no longer in a shared directory, or is empty .. */
			share->status = SHARE_STAT_DEL;
			del_shares++;
		}
		else if(st.st_mtime != share->lastmod)
		{
			/* file has been modified */
			share->status = SHARE_STAT_UPD;
			upd_shares++;
		}
		else
		{
			switch(share->status)
			{
			case SHARE_STAT_OLD:
				old_shares++;
				break;
			case SHARE_STAT_NEW:
				new_shares++;
				break;
			case SHARE_STAT_UPD:
				upd_shares++;
				break;
			case SHARE_STAT_DEL:
				del_shares++;
				break;
			}
		}

		DEBUG("Loaded share from database: %s/%s", share->path, share->file);
	}

	if(err)
	{
		share_del(share);
		share = NULL;
		NOTICE("Error while reading sharedb: %s", ERRMSG());
	}

	fclose(fd);

	return !err;
}


/* adds a directory to the sharelist */
void share_dir(unsigned char *path, int depth)
{
	struct sharestruct *share;
	DIR *dir;
	struct dirent *dirp;
	struct stat st;
	unsigned char rpath[MAX_PATH];
	int ftype;

	context;

	if(!running)
		return;

	/* store the current path */
	if(depth == 0)
		curpath = path;

	/* don't nest too deep */
	if(depth > MAX_DEPTH)
	{
		NOTICE("Nesting too deep!\n");
		return;
	}

	if(!(dir = opendir(path)))
	{
		NOTICE("Cannot read directory '%s'", path);
		return;
	}

	while((dirp = readdir(dir)))
	{
		/* dont read hidden files, ".", or ".." */
		if(dirp->d_name[0] == '.')
			continue;

		context;

		/* full path */
		if(snprintf(rpath, sizeof(rpath), "%s/%s", path, dirp->d_name) == -1)
		{
			NOTICE("Path too long!");
			continue;
		}

		/* stat */
		if(stat(rpath, &st))
		{
			NOTICE("Couldn't stat file '%s': %s", rpath, ERRMSG());
			continue;
		}

		/* don't share empty files - the servers hate those ;) */
		if(st.st_size == 0)
			continue;

		/* don't read links */
		if(S_ISLNK(st.st_mode))
			continue;

		/* recurse */
		if(S_ISDIR(st.st_mode))
		{
			share_dir(rpath, depth+1);
			continue;
		}

		ftype = get_file_type(dirp->d_name);
		if(ftype == SHARE_TYPE_UNKNOWN)
			continue;

		context;

		if(is_shared(path, dirp->d_name))
			continue;

		share = share_new();
		if(share)
		{
			/* add new share */
			new_shares++;
			share->path = ag_strdup(path);
			share->file = ag_strdup(dirp->d_name);
			share->hash = share_make_hash(path, dirp->d_name);

			share->size = st.st_size;
			share->lastmod = st.st_mtime;
			share->status = SHARE_STAT_NEW;

			DEBUG("Read new share: %s/%s", share->path, share->file);
		}
		else
		{
			NOTICE("Error allocating new share: %s", ERRMSG());
		}
	}

	closedir(dir);
}


void share_free()
{
	struct sharestruct *share=sharelist,*next;

	context;

	while(share)
	{
		next = share->next;
		share_del(share);
		share = next;
	}
	sharelist = NULL;
}


void share_list(struct confshare *s)
{
	while(s)
	{
		NOTICE("Reading %s ...", s->path);
		share_dir(s->path, 0);
		s = s->next;
	}
}

